#!/usr/bin/env node
'use strict';

import childProcess from 'node:child_process';
import fs from 'node:fs';
import os from 'node:os';
import path from 'node:path';
import { Logger } from "./jsview-logger.js";

function printArgumentsUsages(requiredUsages, optionalUsages, helpUsages)
{
  let maxLength = 0;
  for (const usage in requiredUsages) {
    maxLength = (maxLength > usage.length ? maxLength : usage.length);
  }
  for (const usage in optionalUsages) {
    maxLength = (maxLength > usage.length ? maxLength : usage.length);
  }
  for (const usage in helpUsages) {
    maxLength = (maxLength > usage.length ? maxLength : usage.length);
  }

  Logger.Info('Usages:');
  for (const usage in requiredUsages) {
    const spaceLength = maxLength - usage.length;
    Logger.Info();
    Logger.Info('  ' + usage + ':  ' + ' '.repeat(spaceLength) + '[Required] ' + requiredUsages[usage]);
  }
  for (const usage in optionalUsages) {
    const spaceLength = maxLength - usage.length;
    Logger.Info();
    Logger.Info('  ' + usage + ':  ' + ' '.repeat(spaceLength) + '[Optional] ' + optionalUsages[usage]);
  }
  for (const usage in helpUsages) {
    const spaceLength = maxLength - usage.length;
    Logger.Info();
    Logger.Info('  ' + usage + ':  ' + ' '.repeat(spaceLength) + helpUsages[usage]);
  }
  Logger.Info();
}

function parseArguments(requiredUsages = {},
                        optionalUsages = {},
                        withoutUnparsed = true)
{
  const keySeparator = '|';
  const formatKey = (key) => {
    let formattedKey = key.trim().replace(/^-*/, '');
    formattedKey = formattedKey.replace(/-(.)/g, (x) => x[1].toUpperCase())
    return formattedKey;
  };
  const formatValue = (value) => {
    let formattedValue = (typeof(value) === 'string' ? value.trim() : value);
    if (value === undefined
    || value === null
    || value === 'undefined'
    || value === 'null'
    || value === '') {
      formattedValue = null;
    }
    return formattedValue;
  }
  const hasKey = (usages, key) => {
    const fmtKey = formatKey(key);
    for (const checker in usages) {
      const checkerList = checker.split(keySeparator);
      for (const alias of checkerList) {
        const fmtAlias = formatKey(alias);
        if (fmtAlias === fmtKey) {
          return true;
        }
      }
    }

    return false;
  }
  const checkAliases = (checker, options) => {
    const checkerList = checker.split(keySeparator);

    const formattedAliases = [];
    for (const alias of checkerList) {
      const fmtAlias = formatKey(alias);
      formattedAliases.push(fmtAlias);
    }

    let aliasValue = undefined;
    let found = false;
    for (const alias of formattedAliases) {
      if (alias in options) {
        aliasValue = options[alias];
        found = true;
        break;
      }
    }

    if (found == false) {
      return null;
    }

    const aliasedOptions = {};
    for (const alias of formattedAliases) {
      aliasedOptions[alias] = aliasValue;
    }

    return aliasedOptions;
  };

  const helpUsages = {
    '-h | --help': 'Print help usages.',
  }

  let options = {
    unparsed: []
  };

  // arg0: node
  // arg1: script.js
  const arguList = process.argv.slice(2);
  for (const argu of arguList) {
    let [key, value] = argu.split('=');
    if (argu.includes('=') == false) {
      value = true;
    }
    if (key.startsWith('-') == false) {
      if (withoutUnparsed) {
        Logger.ErrorAndExitNoException('Failed to parse argument: ' + argu);
      } else {
        options.unparsed.push(key);
        continue;
      }
    }

    if (hasKey(helpUsages, key) == false
    && hasKey(requiredUsages, key) == false
    && hasKey(optionalUsages, key) == false) {
        Logger.ErrorAndExitNoException('Failed to parse usage: ' + argu);
    }

    let formattedKey = formatKey(key);
    let formattedValue = formatValue(value);
    options[formattedKey] = formattedValue;
  }

  for (const checker in helpUsages) {
    const aliasedOptions = checkAliases(checker, options);
    if (aliasedOptions) {
      printArgumentsUsages(requiredUsages, optionalUsages, helpUsages);
      process.exit(0);
    }
  }
  for (const checker in requiredUsages) {
    const aliasedOptions = checkAliases(checker, options);
    if (!aliasedOptions) {
      Logger.ErrorAndExitNoException('Failed to find required option: ' + checker);
    }
    options = { ...options, ...aliasedOptions };
  }
  for (const checker in optionalUsages) {
    const aliasedOptions = checkAliases(checker, options);
    options = { ...options, ...aliasedOptions };
  }

  return options;
}

/**************************************************
 * jsview-test-project
 * ├── dist
 * │   ├── debug
 * │   │   └── map
 * │   └── js
 * ├── node_modules
 * │   ├── .vite
 * │   │   └── jsview
 * │   ├── @shijiu
 * │   │   ├── jsview -> ../../../../@shijiu/jsview
 * │   │   │   ├── dom
 * │   │   │   │   ├── bin
 * │   │   │   │   ├── code
 * │   │   │   ├── loader
 * │   │   │   ├── patches
 * │   │   │   │   └── node_modules
 * │   │   │   ├── tools
 * │   │   ├── jsview-vue -> ../../../../@shijiu/jsview-vue
 * │   │   │   └── utils
 * │   │   └── jsview-vue-samples -> ../../../../@shijiu/jsview-vue-samples
 * └── src
 *     └── appConfig
*************************************************/
function getOptions(framework)
{
  const options = {};

  options.projectDir = process.cwd();

  options.modulesDir = path.resolve(options.projectDir, 'node_modules');
  options.jsviewDir = path.resolve(options.modulesDir, '@shijiu', 'jsview');
  options.jsviewDomDir = path.resolve(options.jsviewDir, 'dom');
  options.jsviewPatchesDir = path.resolve(options.jsviewDir, 'patches');
  options.jsviewPatchModulesDir = path.resolve(options.jsviewPatchesDir, 'node_modules');
  options.jsviewToolsDir = path.resolve(options.jsviewDir, 'tools');
  options.jsviewDomRevisionFile = path.resolve(options.jsviewDir, 'dom', 'target_core_revision.mjs');

  options.jsviewVueDir = path.resolve(options.modulesDir, '@shijiu', 'jsview-vue');
  options.jsviewReactDir = path.resolve(options.modulesDir, '@shijiu', 'jsview-react');
  if (framework) {
    options.jsviewFrameworkDir = path.resolve(options.modulesDir, '@shijiu', 'jsview-' + framework);
  }

  options.appConfigDir = path.resolve(options.projectDir, 'src', 'appConfig');
  options.appPrivKeyFile = path.resolve(options.appConfigDir, 'app_sign_private_key.crt');
  options.appPubKeyFile = path.resolve(options.appConfigDir, 'app_sign_public_key.pem');

  options.distDir = path.resolve(options.projectDir, 'dist');
  if (framework == 'vue') {
    const vueConfigFile = path.resolve(options.projectDir, 'vue.config.js');
    if (fs.existsSync(vueConfigFile)) {
      let vueCfgContent = fs.readFileSync(vueConfigFile, 'utf8');
      const vueCfgObj = JSON.parse(vueCfgContent);
      if (vueCfgObj.outputDir) {
        options.distDir = path.resolve(options.projectDir, vueCfgObj.outputDir);
      }
    }
  }
  options.distJsvListFile = path.resolve(options.distDir, 'jsv-list.json');
  options.distJsvInfoFile = path.resolve(options.distDir, 'jsv-info.json');
  options.distJsDir = path.resolve(options.distDir, 'js');
  options.distDebugDir = path.resolve(options.distDir, 'debug');
  options.distDebugMapDir = path.resolve(options.distDebugDir, 'map');

  const cacheName = (framework == 'vue' ? '.vite' : '.cache');
  options.cacheDir = path.resolve(options.modulesDir, cacheName);
  options.baseUrlFile = path.resolve(options.cacheDir, 'jsview', 'network.mjs');

  return options;
}

function getPackageObject(modulePath)
{
  const pkgFullFile = path.resolve(modulePath, 'package.json');
  if (!fs.existsSync(pkgFullFile)) {
    Logger.ErrorAndExit('Failed to get ' + modulePath + ', file is not exists.');
  }
  let pkgContent = fs.readFileSync(pkgFullFile, 'utf8');
  const pkgObj = JSON.parse(pkgContent);

  return pkgObj;
}

function cpSync(workDir, srcPath, destPath, ignore=[])
{
  const exists = fs.existsSync(srcPath);
  const stats = exists && fs.statSync(srcPath);
  const isDirectory = exists && stats.isDirectory();
  if (isDirectory) {
    const childFileNames = fs.readdirSync(srcPath);
    for (const childName of childFileNames) {
      if (fs.existsSync(destPath) == false) {
        fs.mkdirSync(destPath, { recursive: true });
      }
      cpSync(
        workDir,
        path.join(srcPath, childName),
        path.join(destPath, childName),
        ignore
      );
    };
  } else {
    const filename = path.basename(srcPath);
    if (ignore && ignore.includes(filename)) {
      return;
    }
    const desc = ' Copying ' + path.relative(workDir, srcPath) + ' => ' + path.relative(workDir, destPath);
    Logger.Info(desc);
    const dirPath = path.dirname(destPath);
    if (fs.existsSync(dirPath) == false) {
      fs.mkdirSync(dirPath, { recursive: true });
    }
    fs.copyFileSync(srcPath, destPath);
  }
}

function rmSync(workDir, filePath)
{
  const desc = '  Removing ' + path.relative(workDir, filePath);
  Logger.Info(desc);
  fs.rmSync(filePath, { recursive: true, force: true });
}

function symlinkSync(workDir, targetPath, toPath)
{
  fs.rmSync(toPath, { recursive: true });

  const relativePath = path.relative(path.dirname(toPath), targetPath);

  const desc = '  ' + path.relative(workDir, toPath) + ' Linking ' + path.relative(workDir, relativePath);
  Logger.Info(desc);
  fs.symlinkSync(relativePath, toPath, 'junction');
}

function isSymlinkSync(filePath)
{
  const stats = fs.lstatSync(filePath);

  return stats.isSymbolicLink();
}

function execCommand(cmdline, withOutput)
{
  Logger.Info('Run [' + cmdline + ']... ');
  try {
    const ret = childProcess.execSync(cmdline, { stdio: withOutput ? 'pipe' : 'inherit' });
    if (ret) {
      return ret.toString();
    }
    return null;
  } catch (err) {
    if (withOutput) {
      return err.stderr?.toString()
    } else {
      Logger.ErrorAndExit(err)
    }
  }
}

function checkNodeVersion(minMajorVersion = 16)
{
  const versions = process.version.replace('v', '').split('.');
  const [majorVersion, minorVersion, patchVersion] = versions;

  Logger.Info(`Node version: v${majorVersion}.${minorVersion}.${patchVersion}`);
  if (majorVersion < minMajorVersion) {
    Logger.ErrorAndExit('Node version is too low! Please upgrade your Node.js first.');
  }
}

function askAndAnswer(question)
{
  process.stdout.write(question)
  let buffer = Buffer.alloc(100)
  fs.readSync(0, buffer, 0, 100)
  let answer = buffer.toString().toLowerCase().replace('\n', '');
  answer = answer.substring(0, answer.indexOf('\0'));
  return answer;
}

function readJsonFile(filePath) {
  try {
    const pkgJsonContent = fs.readFileSync(filePath, 'utf8');
    const pkgJsonObj = JSON.parse(pkgJsonContent);

    return pkgJsonObj;
  } catch (ex) {
  }

  return null;
}

function makeZip(zipFilePath, password, ...args)
{
  zipFilePath = path.resolve(zipFilePath);
  const zipFileDir = path.dirname(zipFilePath);
  const zipFileName = path.basename(zipFilePath);

  //  创建一个空zip包
  let cmdLine = 'cd ' + zipFileDir + ' && rm -f ' + zipFileName;
  cmdLine += ' && echo "dump" |zip ' + zipFileName + ' -';
  cmdLine += ' && zip -d ' + zipFileName + ' "*"';
  try {
    childProcess.execSync(cmdLine);
  } catch (error) {
    Logger.ErrorAndExit('Failed to create zip file ' + zipFilePath, error);
  }

  for(const source of args) {
    let cmdLine = 'cd ' + path.dirname(source)
    cmdLine += ' && zip -ru ' + (password ? `-P ${password} ` : '') + zipFilePath + ' ' + path.basename(source);
    try {
      childProcess.execSync(cmdLine);
      Logger.Info('Success to append file ' + source);
    } catch (error) {
      Logger.ErrorAndExit('Failed to make zip file from ' + source, error);
    }
  }
}

function convertHomeDirectory(filePath)
{
  if(!filePath) {
    return null;
  }

  const homeDir = os.homedir();
  return filePath.replace(/^~\//, homeDir + '/');
}

function getNpmRegistry()
{
  return 'https://nexus.cluster.qcast.cn/repository/npm-public';
}

export {
  askAndAnswer,
  checkNodeVersion,
  convertHomeDirectory,
  cpSync,
  execCommand,
  getOptions,
  getPackageObject,
  getNpmRegistry,
  isSymlinkSync,
  makeZip,
  parseArguments,
  readJsonFile,
  rmSync,
  symlinkSync,
  Logger,
}