#!/usr/bin/env node
'use strict';

import fs from 'node:fs';
import path from 'node:path';
import {
    askAndAnswer,
    checkNodeVersion,
    convertHomeDirectory,
    execCommand,
    getOptions,
    parseArguments,
    Logger,
} from './jsview-common.mjs';

function getDefaultSourceMapDir() {
  return './dist/debug/map';
}

function getExtraOptions(argv)
{
  const options = getOptions(argv.framework);

  options.inputFile = convertHomeDirectory(argv.inputFile);
  options.outputFile = convertHomeDirectory(argv.outputFile);
  options.sourcemapDir = convertHomeDirectory(argv.sourcemapDir) ?? getDefaultSourceMapDir();
  options.sourceMapMoudulePath = path.resolve(options.modulesDir, 'source-map', 'source-map.js');

  if(!argv.sourcemapDir) {
    Logger.Warn('Use default source map dir: ' + options.sourcemapDir);
  }
  if(fs.existsSync(options.sourcemapDir) == false) {
    Logger.ErrorAndExit('Source map dir is not exists. Please check dir: ' + options.sourcemapDir);
  }

  return options;
}

async function ensureSourceMapModule(options)
{
  if (fs.existsSync(options.sourceMapMoudulePath)) {
    return;
  }

  Logger.Warn('node module "source-map" is not found.')
  Logger.Warn(' Do you want to install it?');
  const answer = askAndAnswer('(Y/n) ');
  if (answer && answer.substring(0, 1) != 'y' && answer.substring(0, 3) != 'yes') {
    Logger.ErrorAndExitNoException('User cancelled.')
  }

  let cmdline = 'npm install --package-lock-only --save-exact --save-dev source-map';
  execCommand(cmdline);
  cmdline = 'npm ci'
  execCommand(cmdline);
}

async function parseConsoleLog(options)
{
  options.sourceMapModule = await import(options.sourceMapMoudulePath);
  options.sourceMapConsumerMap = { };

  const content = fs.readFileSync(options.inputFile, 'utf-8');
  const contentLines = content.split('\n');

  const regexCharAt = /^\s+at\s+(\S+)?\s?\(?(.*):(\d+):(\d+)\)?$/;
  const regexSymbolAt = /^(\S+) @ (.*?):(\d+)(?::(\d+))?$/;

  var resultLines = [];
  for(const line of contentLines) {
    var recovered = line.replace(/.*?:\d+ /, '');

    if(regexCharAt.test(recovered)) {
      recovered = await recoverCharAt(options, recovered, regexCharAt);
    } else if(regexSymbolAt.test(recovered)) {
      recovered = await recoverSymbolAt(options, recovered, regexSymbolAt);
    }

    if(!options.outputFile) {
      console.log(recovered);
    } else {
      resultLines.push(recovered);
    }
  }

  const result = resultLines.join('\n');
  if(options.outputFile) {
    fs.writeFileSync(options.outputFile, result, 'utf-8');
  }
}

async function recoverCharAt(options, content, regex)
{
  const matchResult = regex.exec(content);
  var funcName = matchResult?.[1];
  var fileName = matchResult?.[2];
  var fileRow = matchResult?.[3];
  var fileColumn = matchResult?.[4];

  // at file.js:row:column
  if(!fileName) {
    fileName = funcName;
    funcName = "";
  } else { // at func (file.js:row:column)
    // do nothing
  }

  const origPos = await getOriginalPosition(options, fileName, fileRow, fileColumn);
  if(!origPos) {
    return content;
  }

  funcName = (origPos.name ?? funcName);
  content  = '    at ';
  content += funcName
  content += (funcName ? ' (' : '');
  content += (origPos.source + ':' + origPos.line + ':' + origPos.column);
  content += (funcName ? ') ' : '');

  return content;
}

async function recoverSymbolAt(options, content, regex)
{
  // 忽略，拿不到column，不能解析出结果
  // const matchResult = regex.exec(content);
  // var funcName = matchResult?.[1];
  // var fileName = matchResult?.[2];
  // var fileRow = matchResult?.[3];
  // var fileColumn = matchResult?.[4];

  // const origPos = await getOriginalPosition(options, fileName, fileRow, fileColumn);
  // if(!origPos) {
  //   return content;
  // }

  // funcName = (origPos.name ?? funcName);
  // content = funcName
  // content += ' @ '
  // content += (origPos.source + ':' + origPos.line + ':' + origPos.column);

  return content;
}

async function getOriginalPosition(options, fileName, fileRow, fileColumn)
{
  if(!fileName || !fileRow || !fileColumn) {
    return null;
  }

  const sourceMapFilePath = path.resolve(options.sourcemapDir, fileName + '.map');
  if(fs.existsSync(sourceMapFilePath) == false) {
    Logger.Warn('Source map file is not found: ' + sourceMapFilePath)
    return null;
  }

  var sourceMapConsumer = options.sourceMapConsumerMap[sourceMapFilePath];
  if(!sourceMapConsumer) {
    const sourceMapContent = fs.readFileSync(sourceMapFilePath, 'utf-8');
    sourceMapConsumer = new options.sourceMapModule.SourceMapConsumer(sourceMapContent);
	  if(sourceMapConsumer instanceof Promise) {
	    sourceMapConsumer = await sourceMapConsumer ;
	  }
    options.sourceMapConsumerMap[sourceMapFilePath] = sourceMapConsumer;
  }

  const originalPosition = sourceMapConsumer.originalPositionFor({
    line: Number(fileRow),
    column: Number(fileColumn),
  });

  return originalPosition;
}

async function main(argv)
{
  checkNodeVersion();

  const options = getExtraOptions(argv);

  await ensureSourceMapModule(options);

  await parseConsoleLog(options);
}

const requiredUsages = {
  '--input-file': 'Chromium console log file path.',
};
const optionalUsages = {
  '-d | --sourcemap-dir': 'Source map dir. Default dir is: ' + getDefaultSourceMapDir(),
  '-o | --output-file': 'Recovered console log file path.',
};
const argv = parseArguments(requiredUsages, optionalUsages, false);
main(argv)
